/* $Header: /CVSROOT/public/scylla-charybdis/md5backup/md5backup.c,v 1.18 2004/07/25 09:16:41 tino Exp $
 *
 * Copyright (C)2003,2004 by Valentin Hilbig
 *
 * MD5 based harddrive to harddrive backup utility.
 * Compared to my other utilities it is incredible fast,
 * as it is designed for speed.
 *
 * It should run over NFS too, but this never has been tested.
 *
 * This is release early code.  Use at own risk.
 *
 * $Log: md5backup.c,v $
 * Revision 1.18  2004/07/25 09:16:41  tino
 * see ChangeLog: bugfixes in bin/dobackup.sh and tinolib
 *
 * Revision 1.17  2004/07/05 23:53:52  tino
 * va_copy not defined in all systems
 *
 * Revision 1.16  2004/06/27 15:17:31  tino
 * removed superfluous ignore and printf
 *
 * Revision 1.15  2004/06/19 12:31:29  tino
 * minor improvements
 *
 * Revision 1.14  2004/06/18 23:51:16  tino
 * see ChangeLog
 *
 * Revision 1.13  2004/06/12 11:17:46  tino
 * removed too frequent printing of unneccessary warnings
 *
 * Revision 1.12  2004/05/23 10:58:35  tino
 * once again, access to (same) feed pointer.  Now globally fixed I hope.
 *
 * Revision 1.11  2004/05/23 07:55:45  tino
 * harmless bug fixed: access to already freed memory
 *
 * Revision 1.10  2004/05/21 03:50:09  tino
 * known files are now linked into linked_file_store, too.
 * Needed to add warnings, as currently this solution is not perfect.
 *
 * Revision 1.9  2004/05/04 17:51:47  tino
 * O_DIRECT does not work, fix does not work either, so it's disabled now
 *
 * Revision 1.8  2004/05/01 04:37:36  tino
 * now it works as expected
 *
 * Revision 1.7  2004/05/01 02:15:01  tino
 * New feature: =/PATH/
 *
 * Revision 1.6  2004/01/12 03:55:02  tino
 * Includes, excludes added but not tested.
 *
 * Revision 1.5  2004/01/11 22:26:08  tino
 * Version and compile date in usage now
 *
 * Revision 1.4  2004/01/11 22:04:32  tino
 * Some last bugs fixed
 *
 * Revision 1.3  2004/01/11 20:29:29  tino
 * Release candidate: Added features:
 * Autoignore and stay in local filesystem
 *
 * Revision 1.2  2004/01/07 12:05:09  tino
 * Scripts improved, some spelling corrections, and CVS headers added.
 * md5backup.c now has a sanity check which slows down things factor 10.
 */
#define _GNU_SOURCE	1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <setjmp.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/stat.h>
#include <sys/time.h>

#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/fcntl.h>

#include "tino/debug.h"
#include "tino/alloc.h"
#include "tino/slist.h"
#include "tino/str.h"
#include "tino/file.h"
#include "tino/sysfix.h"

#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "md5.h"

#include <gdbm.h>

#include <sys/types.h>
#include <dirent.h>

#include "version.h"

/* I added checks because I thought they are useful, but I was wrong:
 * These checks slow down the process by factors (10 or more)!
 * You really don't want and need this.  Trust me ;)
 * Instead:
 * From while to while and in case check.sh finds errors
 * just delete the the database to get a fresh full backup.
 * (FYI the database is "/target/path/"dbm/"nodename")
 */
#undef	MD5BACKUP_SANITY_CHECKS

static struct
  {
    /* Runtime handles
     */
    GDBM_FILE	db;
    FILE	*log, *md5;

    /* Runtime global counts
     */
    time_t	start, last;
    int		targets, ignored;
    int		errors, warnings, dirs, examined, changed, known, added;
    unsigned long long	copy;

    /* Runtime information
     */
    int		target_length;
    const char	*target;

    /* Runtime buffers
     */
    TINO_SLIST	dirlist;
    TINO_SLIST	linkpath_list;
    TINO_GLIST	ignore_list_dir, ignore_list_file;
    TINO_GLIST	target_list;
    size_t	bigbuf_max;
    void	*bigbuf;
    int		o_direct;


    /* Basic config
     */
    const char	*path, *node;
    int		width;

    /* Calculated config
     */
    char	dbfile[PATH_MAX];
    char	outpath[PATH_MAX], tmpfile[PATH_MAX];
    char	logfile[PATH_MAX], md5file[PATH_MAX];
  } config;

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* Console output stuff	*/	/* shall use tinolib here	*/

static void
progress_clear(void)
{
  /* Should only be send to a tty
   */
  printf("%*s\r", config.width, "");
}

static void
progress(void)
{
  char		buf[200];
  int		n;
  time_t	now;

  time(&now);
  now	-= config.start;
  n	= snprintf(buf, sizeof buf,
		   "%02d:%02d %d t%d i%d e%d w%d d%d f%d m%d o%d n%d %LdMB ",
		   (int)(now/60),
		   (int)(now%60),
		   (config.dirlist ? config.dirlist->count : 0),
		   config.targets,
		   config.ignored,
		   config.errors,
		   config.warnings,
		   config.dirs,
		   config.examined,
		   config.changed,
		   config.known,
		   config.added,
		   config.copy>>20);
  fputs(buf, stdout);
  n	= config.target_length-config.width+n;
  if (n<0)
    n	= 0;
  if (config.target && n<config.target_length)
    fputs(config.target+n, stdout);
  if (n+config.target_length<config.width)
    putchar(' ');
  putchar('\r');
  fflush(stdout);
}

static void
vlog(const char *prefix, const char *s, va_list list, const char *suffix, ...)
{
  time_t	tim;
  struct tm	tm;
  va_list	tmp;

  time(&tim);
  if (config.last!=tim)
    progress();
  config.last	= tim;
  gmtime_r(&tim, &tm);

  if (!config.log)
    return;
  fprintf(config.log,
	  "%4d-%02d-%02d %02d:%02d:%02d %s",
	  1900+tm.tm_year, tm.tm_mon+1, tm.tm_mday,
	  tm.tm_hour, tm.tm_min, tm.tm_sec, prefix);
  vfprintf(config.log, s, list);
  va_start(tmp, suffix);
  vfprintf(config.log, suffix, tmp);
  va_end(tmp);
  fputc('\n', config.log);
  fflush(config.log);
}

static void
log(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  vlog("", s, list, "");
  va_end(list);
}

static void
verror(const char *prefix, const char *s, va_list list)
{
  const char	*err;
  va_list	list2;

  err	= strerror(errno);

  config.errors++;

  TINO_VA_COPY(list2, list);
  vlog(prefix, s, list, ": %s", err);
  va_end(list2);

  fputs(prefix, stderr);
  vfprintf(stderr, s, list);
  fprintf(stderr, ": %s\n", err);
}

static void
error(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  verror("error: ", s, list);
  va_end(list);
}

static void
fatal(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  verror("fatal: ", s, list);
  va_end(list);
  exit(-1);
  abort();
  for(;;);
}

static void
vwarn(const char *s, va_list list)
{
  va_list	list2;

  config.warnings++;

  TINO_VA_COPY(list2, list);
  vlog("warning: ", s, list, "");
  va_end(list2);

  fputs("warning: ", stderr);
  vfprintf(stderr, s, list);
  fprintf(stderr, "\n");
}

static void
warn(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  vwarn(s, list);
  va_end(list);
}



/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* helpers	*/	/* shall use tinolib here	*/

static char *
assemble(char *s, size_t max, const char *fmt, ...)
{
  int		n;
  va_list	list;

  va_start(list, fmt);
  n	= vsnprintf(s, max, fmt, list);
  va_end(list);
  if (n>=max)
    fatal("argument too long");
  return s;
}

#ifdef MD5BACKUP_SANITY_CHECKS
static int
existsfile(const char *file)
{
  struct stat64		st;

  return (!lstat64(file, &st) && S_ISREG(st.st_mode));
}
#endif

static int
existsdir(const char *dir)
{
  struct stat	s;

  if (stat(dir, &s))
    return 0;
  return S_ISDIR(s.st_mode);
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* Directory handling	*/	/* Shall become tinolib	*/

struct sc_dir_stat
  {
    const char		*name;	/* Full path	*/
    struct stat64	st;
  };

/* This gobbles CWD!
 */
static TINO_GLIST
sc_scandir(const char *name)
{
  TINO_SLIST	slist;
  TINO_GLIST	glist;
  DIR		*dir;
  struct dirent	*dp;
  char		buf[PATH_MAX+2];
  const char	*s;

  xDP(("sc_scandir(%s)", name));
  /* Read directory contents
   */
  if ((dir=opendir(name))==NULL)
    {
      error("cannot open dir: %s", name);
      return 0;
    }
  slist	= tino_slist_new();
  while ((dp=readdir(dir))!=NULL)
    if (!(dp->d_name[0]=='.' &&
	  (dp->d_name[1]==0 ||
	   (dp->d_name[1]=='.' && dp->d_name[2]==0))))
	{
	  xDP(("sc_scandir() readdir %s", dp->d_name));
	  tino_slist_add(slist, dp->d_name);
	}
  if (closedir(dir))
    error("error reading dir: %s", name);

  /* Now do the stats
   */
  glist	= tino_glist_new(sizeof (struct sc_dir_stat));
  for (; (s=tino_slist_get(slist))!=NULL; tino_slist_free(s))
    {
      struct stat64		st;
      struct sc_dir_stat	*ent;

      xDP(("sc_scandir() stat %s", s));
      tino_strxcpy(buf, name, sizeof buf);
      if (!*buf || buf[strlen(buf)-1]!='/')
	tino_strxcat(buf, "/", sizeof buf);
      tino_strxcat(buf, s, sizeof buf);
      if (strlen(buf)>=PATH_MAX)
	{
	  error("filename too deep: %s/%s", name, s);
	  continue;
	}
      if (lstat64(buf, &st))
	{
	  error("cannot stat: %s", buf);
	  continue;
	}
      ent	= tino_glist_add(glist)->data;
      ent->name	= tino_strdup(buf);
      ent->st	= st;
    }
  xDP(("sc_scandir() ok"));
  return glist;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* Database stuff	*/

struct data
  {
    unsigned long long	mtime;
    int			count;
    char		md5[34];
  };

static int
dbfetch(const char *name, struct data *data)
{
  char	tmp[32+32+32];
  datum	d;

  config.examined++;
  d.dptr	= (char *)name;
  d.dsize	= strlen(name);
  d		= gdbm_fetch(config.db, d);
  if (d.dptr==0)
    return 0;
  xDP(("db(%s) %d '%.*s'", name, d.dsize, d.dsize, d.dptr));
  if (d.dsize<sizeof tmp)
    {
      unsigned long long	tim, cnt;
      int			p1, p2;

      memcpy(tmp, d.dptr, d.dsize);
      free(d.dptr);
      tmp[d.dsize]	= 0;

      p1=0;
      p2=0;
      sscanf(tmp, "%Lu %Lu %n%*s%n", &tim, &cnt, &p1, &p2);
      if (3<p1 && p1+32==p2)
	{
	  data->mtime	= tim;
	  data->count	= (int)cnt;
	  memcpy(data->md5, tmp+p1, 32);
	  data->md5[32]	= 0;
	  return 1;
	}
    }
  error("database corrupt for %s", name);
  return 0;
}

static int
dbstore(const char *name, struct data *data)
{
  char	tmp[32+32+32];
  datum	k,v;

  assemble(tmp, sizeof tmp,
	   "%Lu %Lu %s",
	   (unsigned long long)data->mtime,
	   (unsigned long long)data->count,
	   data->md5);
  k.dptr	= (char *)name;
  k.dsize	= strlen(name);
  v.dptr	= tmp;
  v.dsize	= strlen(tmp);
  if (!gdbm_store(config.db, k, v, GDBM_REPLACE))
    return 1;
  error("database not writeable for %s", name);
  return 0;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* Physical low level IO	*/

static int
readit(int fd, char *buf, size_t len)
{
  int	got;

  while ((got=read(fd, buf, len))<0)
    if (errno!=EINTR && errno!=EAGAIN)
      break;
  return got;
}

static int
writeall(int fd, const char *buf, size_t len)
{
  int	pos, put;

  for (pos=0; pos<len; )
    {
      put	= write(fd, buf+pos, len-pos);
      if (put<=0)
	{
	  if (!put)
	    break;
	  if (errno!=EINTR && errno!=EAGAIN)
	    break;
	  continue;
	}
      if (put>len-pos)
	fatal("syscall defect, wrote more than allowed");
      pos+=put;
    }
  return pos;
}

static int
readcmp(int fd, const char *buf, size_t len)
{
  int	pos, got, max;
  char	blk[BUFSIZ*16];

  for (pos=0; pos<len; pos+=got)
    {
      max	= len-pos;
      if (max>sizeof blk)
	max	= sizeof blk;
      got	= readit(fd, blk, max);
      if (got<=0)
	{
	  if (got)
	    return -1;
	  break;
	}

      if (got>max)
	fatal("syscall defect, read more than allowed");

      if (memcmp(buf+pos, blk, got))
	return 1;
    }
  return 0;
}

/* Compare open FD to backup tmpfile
 *
 * Returns:
 * <0	on error
 * 0	if identical
 * >0	if differs
 */
static int
compare_file(int fd, const char *name)
{
  int	cmp;
  int	res;
  int	got;

  xDP(("compare_file(%d,%s)", fd,name));

  if ((cmp=open(config.tmpfile, O_RDONLY|O_LARGEFILE, 0775))<0)
    fatal("cannot read tmpfile: %s", config.tmpfile);

  res	= 0;
  while ((got=readit(fd, config.bigbuf, config.bigbuf_max))>0)
    if ((res=readcmp(cmp, config.bigbuf, got))!=0)
      {
	if (res<0)
	  fatal("read error: %s", config.tmpfile);
	break;
      }
  if (got<0)
    return -1;
  if (!res)
    {
      char	tmp[BUFSIZ];

      /* Check for EOF on both files
       */
      if ((res=readit(cmp, tmp, sizeof tmp))<0)
	fatal("read error: %s", config.tmpfile);
    }
  if (close(cmp))
    fatal("close error: %s", config.tmpfile);
  return res;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* md5backup helpers	*/

static char *
md5filename(char *out, size_t len, const char *pref, const char *md5, int count)
{
  assemble(out, len,
	   "%s/%c%c/%c%c/%s.%d.md5",
	   pref,
	   md5[0], md5[1],
	   md5[2], md5[3],
	   md5, count);
  return out;
}

#ifdef MD5BACKUP_SANITY_CHECKS
/* Test if md5 file still is present
 */
static int
md5exists(const char *md5, int count)
{
  char	out[PATH_MAX];

  return existsfile(md5filename(out, sizeof out, config.outpath, md5, count));
}
#endif

/* pref and md5 is a little bit redundant as this is contained in out.
 * Perhaps later on we will elliminate it,
 * as this requires "out" to be something md5filename has created.
 * Upper directory is reduced by two / and lower directory the one by one /
 */
static int
md5link(const char *file, const char *out, int logit)
{
  char	tmp[PATH_MAX];

  if (!link(file, out))
    return 0;

  if (errno!=ENOENT)
    return -1;

  /* Reduce arg by 2 slashes
   */
  tino_strrcut(tino_strrcut(tino_strxcpy(tmp, out, sizeof tmp), '/'), '/');

  if (!mkdir(tmp, 0775) && logit)
    log("created directory %s", tmp);
  
  /* Should reduce out by 1 slashes
   */
  tino_strrcut(tino_strxcpy(tmp, out, sizeof tmp), '/');
  if (!mkdir(tmp, 0775) && logit)
    log("created directory %s", tmp);

  return link(file, out);
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* ignore handling	*/

static void
ignore_add(const char *name)
{
  struct stat64	st;

  if (!stat64(name, &st))
    {
      TINO_GLIST_ENT	ent;
      TINO_GLIST	g;

      if (S_ISDIR(st.st_mode))
	g	= config.ignore_list_dir;
      else if (S_ISREG(st.st_mode))
	g	= config.ignore_list_file;
      else
	{
	  warn("unsupported ignore argument: %s", name);
	  return;
	}
      ent	= tino_glist_add(g);
      *(struct stat64 *)ent->data	= st;
    }
  else
    warn("cannot stat ignore argument: %s", name);
}

static void
ignore_add2(const char *path, const char *name)
{
  char	tmp[PATH_MAX];

  ignore_add(assemble(tmp, sizeof tmp, "%s%s", path, name));
}

static int
is_ignored(TINO_GLIST list,
	   const char *name, const struct stat64 *st1,
	   const char *type)
{
  TINO_GLIST_ENT	ent;
  dev_t			dev	= st1->st_dev;
  ino_t			ino	= st1->st_ino;

  for (ent=list->list; ent; ent=ent->next)
    {
      struct stat64	*st2;

      st2	= ent->data;
      if (dev==st2->st_dev && ino==st2->st_ino)
	{
	  progress_clear();
	  printf("ignore %s: %s\n", type, name);
	  config.ignored++;
	  log("ignore %s %s", type, name);
	  return 1;
	}
    }
  return 0;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* linkpath handling (linked file store)	*/

struct linkpath_user
  {
    jmp_buf	jmp;
    const char	*md5;
    int		count;
    const char	*file;
    int		force;
  };

/* add a linkpath (linked file store)
 * where new files are hardlinked to via the iterate-function below
 */
static void
linkpath_add(const char *path)
{
  if (!existsdir(path))
    fatal("hardlink directory does not exist: %s", path);
  ignore_add(path);
  tino_slist_add(config.linkpath_list, path);
}

static void
linkpath_iterate_fn(const char *path, void *usr /* struct linkpath_user * */)
{
  struct linkpath_user	*u=usr;
  char			tmp[PATH_MAX];

  md5filename(tmp, sizeof tmp, path, u->md5, u->count);
  if (md5link(u->file, tmp, 0))
    {
      if (errno!=EEXIST)
	fatal("cannot hardlink to linked file store: %s", tmp);
      /* File exists.
       * If we must hardlink, then we have to look for another counter.
       * Not that I expect that this bend will ever be taken!
       */
      if (u->force)
	longjmp(u->jmp, 1);
      /* We are in the linked file store where an old file still
       * exists.  Warn if the files are not hardlinked.
       * Perhaps elliminate the longjmp above later, too,
       * however this would break the iteration in situations,
       * I don't want to break it for today.
       */
      if (tino_file_lstat_diff(tmp, u->file))
	warn("existing mismatch in file store: %s", tmp);
    }
}

/* Link file to the file store under an md5 sum and count.
 *
 * force!=0: return error if destination exists.  Actually this is a
 * bug workaround, it should be the case, that we can detect if a
 * hardlink fails because both files are identical.  If not we have an
 * error (always), else we can silently assume the hardlink is ok.
 *
 * return 0 on success, !=0 otherwise
 */
static int
linkpath_run(const char *file, const char *md5, int count, int force)
{
  struct linkpath_user	u;

  u.md5		= md5;
  u.count	= count;
  u.file	= file;
  u.force	= force;
  if (setjmp(u.jmp))
    return -1;
  tino_slist_iterate(config.linkpath_list, linkpath_iterate_fn, (void *)&u);
  return 0;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/* md5backup specific	*/

/* Do a copy from source to tmpfile
 * with an md5 sum in between.
 */
static int
md5copy(const char *name, char *md5)
{
  MD5_CTX	ctx;
  char		digest[16];
  int		got, i;
  int		in, out;
  int		errstate;

  /* Cache the output as it's likely we must compare it
   */
  if ((out=open(config.tmpfile, O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE, 0775))<0)
    {
      error("cannot create temporary %s", config.tmpfile);
      return 0;
    }
  /* Don't cache the input as it's not likely read again
   */
  if ((in=open(name, O_RDONLY|O_NOFOLLOW|config.o_direct|O_LARGEFILE))<0)
    {
      close(out);
      error("cannot open %s", name);
      return 0;
    }

  MD5_Init(&ctx);

  errstate	= 0;
  while ((got=readit(in, config.bigbuf, config.bigbuf_max))>0)
    {
      MD5_Update(&ctx, config.bigbuf, got);
      if (writeall(out, config.bigbuf, got)!=got)
	{
	  if (errno==ENOSPC)
	    {
	      unlink(config.tmpfile);
	      errno	= ENOSPC;
	      fatal("disk full at tmpfile: %s", config.tmpfile);
	    }

	  /* Hm .. what kind of nonpermanent reoccurring error?
	   * NFS softmount dead?
	   */
	  error("cannot write to tmpfile: %s", config.tmpfile);
	  errstate	= 1;
	  break;
	}
      config.copy	+= got;
    }
  if (got<0)
    {
      error("read error on input: %s", name);
      errstate	= 1;
    }
  if (close(in))
    {
      error("error closing input: %s", name);
      errstate	= 1;
    }
  if (close(out))
    {
      error("error closing tmpfile: %s", config.tmpfile);
      errstate	= 1;
    }
  if (errstate)
    {
      unlink(config.tmpfile);
      return 0;
    }

  MD5_Final(digest, &ctx);

  for (i=0; i<16; i++)
    {
      *md5++	= "0123456789abcdef"[(digest[i]>>4)&0xf];
      *md5++	= "0123456789abcdef"[(digest[i]>>0)&0xf];
    }
  *md5	= 0;
  return 1;
}

/* Write the md5sum compatible log entry
 * Complex, as md5sum may use escaped syntax.
 */
static void
md5log(const char *name, const char *md5)
{
  int	i;

  /* Escaped syntax on files containing \n or \\
   */
  for (i=0; name[i]; i++)
    if (name[i]=='\\' || name[i]=='\n')
      {
	fprintf(config.md5, "\\%s *", md5);
	for (i=0; name[i]; i++)
	  if (name[i]=='\n')
	    {
	      fputc('\\', config.md5);
	      fputc('n', config.md5);
	      continue;
	    }
	  else
	    {
	      if (name[i]=='\\')
		fputc('\\', config.md5);
	      fputc(name[i], config.md5);
	    }
	fputc('\n', config.md5);
      }
  /* Ordinary output
   */
  fprintf(config.md5, "%s *%s\n", md5, name);
}

/* Returns:
 * <0	hardlink failed or retryable error
 * ==0	mismatch (funny, I would rather suspect a corrupt file system)
 * >0	match or successful hardlink done
 */
static int
compare_or_link(const char *md5, int count)
{
  char	out[PATH_MAX];
  int	in;

  /* XXX FIXME XXX
   * Check for compressed here in future
   */
  000;

  xDP(("compare_or_link(%s,%d)", md5,count));

  md5filename(out, sizeof out, config.outpath, md5, count);

  if ((in=open(out, O_RDONLY|O_NOFOLLOW|config.o_direct|O_LARGEFILE))>=0)
    {
      int	diff;
      
      diff	= compare_file(in, out);

      /* In case of error we really don't want to continue
       * as something must be badly broken on the target.
       */
      if (diff<0)
	fatal("read error in backup file: %s", out);
      if (close(in))
	fatal("close error in backup file: %s", out);

      if (diff)
	return 0;

      /* We now have an already known file.
       * Hardlink it to the linked file store.
       */
      if (linkpath_run(out, md5, count, 0))
	error("cannot dup to linkpath: %s", out);

      config.known++;
      return 1;
    }

  if (errno!=ENOENT)
    {
      /* Quite not funny
       * We hit some unexpected error.
       * Be conservative and don't try to disturb any more
       */
      fatal("unexpected error accessing file: %s", out);
    }

  if (linkpath_run(config.tmpfile, md5, count, 1))
    return -1;

  if (!md5link(config.tmpfile, out, 1))
    {
      config.added++;
      return 1;
    }
  return -1;
}

/* Hunt for existing MD5 files.
 * Or move it yourself.
 */
static int
md5hunt(const char *md5)
{
  int	count;

  xDP(("md5hunt(%s)", md5));
  for (count=0;; count++)
    {
      int	ret, reloop;

      /* According to the NFS documentation
       * the NFS server might go out of business
       * right before telling us that a hardlink was done.
       * So we have to reloop a bit here.
       */
      for (reloop=0; (ret=compare_or_link(md5, count))<0; reloop++)
	if (reloop>4)
	  fatal("hardlink failed: %s %d", md5, count);

      /* We had a match,
       */
      if (ret>0)
	break;
    }
  /* In either case,
   * we do not need the tmpfile any more
   */
  unlink(config.tmpfile);
  return count;
}

/* Do the backup
 */
static void
backup(const char *name, unsigned long long mtime)
{
  struct data	d;
  char		md5[34];

  xDP(("backup(%s,%Lu)", name, mtime));
  if (dbfetch(name, &d) && d.mtime==mtime
#ifdef MD5BACKUP_SANITY_CHECKS
      && md5exists(d.md5, d.count)
#endif
      )
    {
      xDP(("backup() not newer"));
      return;
    }
  config.changed++;
#if 0
  log("backup %s", name);
#endif
  if (md5copy(name, md5))
    {
      int	cnt;
      int	known;

      xDP(("backup() copied %s", md5));
      md5log(name, md5);
      known	= config.known;
      cnt	= md5hunt(md5);
      if (cnt>=0)
	{
	  d.mtime	= mtime;
	  d.count	= cnt;
	  strcpy(d.md5, md5);
	  if (dbstore(name, &d))
	    {
	      if (known==config.known)
		log("md5 %s %d %s", md5, cnt, name);
	      else
		log("known md5 %s %d %s", md5, cnt, name);
	      return;
	    }
	}
    }
  log("failed backup %s", name);
}

static void
dirlist_add(const char *name, const struct stat64 *st)
{
  if (S_ISDIR(st->st_mode))
    {
      if (!is_ignored(config.ignore_list_dir, name, st, "dir"))
	tino_slist_add(config.dirlist, name);
    }
  else if (S_ISREG(st->st_mode))
    {
      if (!is_ignored(config.ignore_list_file, name, st, "file"))
	backup(name, st->st_mtime);
    }
}

/* I don't like this, that stat is part of following routine.
 * However this is proof of concept code,
 * so this is ok for now
 */
static void
md5backup(const struct sc_dir_stat *targ)
{
  const char	*s;
  dev_t		dev;

  xDP(("md5backup(%s)", targ->name));

  config.targets++;
  config.target		= targ->name;
  config.target_length	= strlen(config.target);
  dev			= targ->st.st_dev;

  progress_clear();
  log("start %s %s %s", config.path, config.node, targ->name);

  config.dirlist	= tino_slist_new();
  dirlist_add(targ->name, &targ->st);
  for (; (s=tino_slist_get(config.dirlist))!=0; tino_slist_free(s))
    {
      TINO_GLIST	glist;
      TINO_GLIST_ENT	ent;

      config.dirs++;
      progress();

      glist	= sc_scandir(s);
      xDP(("md5backup() glist=%p", glist));
      while ((ent=tino_glist_get(glist))!=0)
	{
	  struct sc_dir_stat	*ptr=ent->data;

	  xDP(("md5backup() name=%s", ptr->name));
	  if (ptr->st.st_dev==dev)
	    dirlist_add(ptr->name, &ptr->st);
	  free((char *)ptr->name);
	  tino_glist_free(ent);
	}
      xDP(("md5backup() next"));
    }

  config.target	= 0;
}

static void
init(const char *path, const char *node)
{
  time(&config.start);

  config.linkpath_list		= tino_slist_new();
  config.ignore_list_dir	= tino_glist_new(sizeof(struct stat64));
  config.ignore_list_file	= tino_glist_new(sizeof(struct stat64));
  config.target_list		= tino_glist_new(sizeof(struct sc_dir_stat));

  config.path	= path;
  config.node	= node;

  assemble(config.logfile, sizeof config.logfile,
	   "%slog/%s", path, node);
  if ((config.log=fopen(config.logfile, "a"))==0)
    fatal("cannot append to logfile %s", config.logfile);

  log("md5backup version %s compiled %s", MD5BACKUP_VERSION, __DATE__);

  ignore_add(config.logfile);
  ignore_add2(path, "log");

  assemble(config.outpath, sizeof config.outpath,
	   "%sout", path);
  if (!existsdir(config.outpath))
    fatal("output directory does not exist: %s", config.outpath);
  ignore_add(config.outpath);

  assemble(config.tmpfile, sizeof config.tmpfile,
	   "%stmp/%s.%d", path, node, getpid());
  ignore_add2(path, "tmp");

  assemble(config.md5file, sizeof config.md5file,
	   "%smd5/%s", path, node);
  if ((config.md5=fopen(config.md5file, "a"))==0)
    fatal("cannot append to md5sum %s", config.md5file);
  ignore_add(config.md5file);
  ignore_add2(path, "md5");

  assemble(config.dbfile, sizeof config.dbfile,
	   "%sdbm/%s", path, node);
  if ((config.db=gdbm_open(config.dbfile,
			   BUFSIZ, GDBM_WRCREAT|GDBM_SYNC, 0775, NULL))==0)
    fatal("cannot open database: %s", config.dbfile);
  ignore_add(config.dbfile);
  ignore_add2(path, "dbm");

  /* Try to open a buffer suitable for O_DIRECT IO.
   * This must be aligned to a multiple of the sector size.
   * and IO must start on a sector boundary.
   *
   * Actually this is not done here exactly.
   * This code works only for filesystems up to a sector size of
   * the system's page size, which is 4096.
   * I don't know of any FS with a bigger sector size.
   *
   * Question which comes in mind:
   * What happens with sparse files?
   *
   * XXX FIXME XXX it does not work, don't know why
   */
  config.bigbuf_max	= 1<<20;
#if 0	/* does not work, stay away	*/
  config.o_direct	= O_DIRECT;
  if (posix_memalign(&config.bigbuf, (size_t)sysconf(_SC_PAGE_SIZE),
		     config.bigbuf_max))
#endif
    {
      config.bigbuf	= tino_alloc(config.bigbuf_max);
      config.o_direct	= 0;
    }
  config.width		= 79;
}

static void
cleanup(void)
{
  FILE		*tmp;
  time_t	now;

  gdbm_close(config.db);

  if (fclose(config.md5))
    error("close md5sum %s", config.md5file);

  unlink(config.tmpfile);

  config.copy	+= (1<<20)-1;	/* round up	*/
  time(&now);
  now	-= config.start;
  log("ready %d:%02d, args=%d ign=%d err=%d warn=%d dir=%d file=%d mod=%d old=%d new=%d MB=%Ld",

      (int)(now/60),
      (int)(now%60),
      config.targets,
      config.ignored,
      config.errors,
      config.warnings,
      config.dirs,
      config.examined,
      config.changed,
      config.known,
      config.added,
      config.copy>>20);

  progress_clear();
  printf("%02d:%02d args=%d ign=%d err=%d warn=%d dir=%d file=%d mod=%d old=%d new=%d %LdMB\n",
	 (int)(now/60),
	 (int)(now%60),
	 config.targets,
	 config.ignored,
	 config.errors,
         config.warnings,
	 config.dirs,
	 config.examined,
	 config.changed,
	 config.known,
	 config.added,
	 config.copy>>20);

  tmp		= config.log;
  config.log	= 0;
  if (fclose(tmp))
    error("close logfile %s", config.logfile);
}

static void
process_targets(void)
{
  TINO_GLIST_ENT	ent;

  while ((ent=tino_glist_get(config.target_list))!=0)
    {
      struct sc_dir_stat	*ds=ent->data;
      
      md5backup(ds);
      free((char *)ds->name);
      tino_glist_free(ent);
    }
}

static void
target(const char *arg, int level)
{
  FILE			*fd;
  struct stat64		st;
  struct sc_dir_stat	*ent;
  char			*tmp;

  switch (*arg)
    {
    case 0:
    case '#':
      break;

    case '=':
      linkpath_add(arg+1);
      break;

    case '@':
      if (++level>20)
	error("recoursion too deep: %s", arg+1);
      if ((fd=fopen(arg+1, "rt"))==NULL)
	{
	  error("cannot open target list file: %s", arg+1);
	  break;
	}
      tmp	= tino_strdup(arg+1);
      while (fgets(config.bigbuf, config.bigbuf_max, fd))
	{
	  size_t	len;

	  if ((len=strlen(config.bigbuf))>0)
	    {
	      if (((char *)config.bigbuf)[--len]=='\n')
		((char *)config.bigbuf)[len]	= 0;
	      /* This possibly invalidates arg
	       */
	      target(config.bigbuf, level);
	    }
	}
      if (ferror(fd))
	error("close target list file: %s", tmp);
      if (fclose(fd))
	error("close target list file: %s", tmp);
      free(tmp);
      break;

    case '-':
      ignore_add(arg+1);
      break;
      
    default:
      if (stat64(arg, &st))
	{
	  error("cannot stat target: %s", arg);
	  break;
	}
      ent	= tino_glist_add(config.target_list)->data;
      ent->name	= tino_strdup(arg);
      ent->st	= st;
      break;
    }
}

int
main(int argc, char **argv)
{
  int	i;

  if (argc<3)
    {
      printf("Usage: %s /backup/path/ nodename target...\n"
             "\tIncrementally backup files to backup path given.\n"
	     "\tVersion %s compiled %s\n"
	     "\n"
             "\tDon't forget the trailing / in the /backup/path/.\n"
             "\tNodename should be `hostname -f`.\n"
             "\tIf target is directory it is scanned for files.\n"
             "\tDirectory scan will not cross mount points.\n"
             "\tIf target is -path this is an (global) exclude.\n"
             "\tIf target is @file this file is read for targets.\n"
             "\tIf target is =path all newly created \"out/...\" files\n"
             "\twill be hardlinked there before they are moved to out/.\n"
             "\tThe sub-structure below this path is arbitrary!\n"
	     "\n"
	     "\tBackup structure below /backup/path/ (quick info):\n"
	     "\tdbm/nodename\tDBM file (not needed for restore)\n"
	     "\tlog/nodename\t(*) Logfile of action\n"
	     "\tmd5/nodename\t(*) File suitable for md5sum --check\n"
	     "\tout/... \t(*) Files are stored under their content's MD5 sum.\n"
	     "\ttmp/    \tTemporary directory, should be empty\n"
             "\t(*) needed for restore (not programmed yet!)\n"
	     , argv[0], MD5BACKUP_VERSION, __DATE__);
      return 1;
    }
  init(argv[1], argv[2]);
  for (i=3; i<argc; i++)
    target(argv[i], 0);
  process_targets();
  cleanup();
  return 0;
}
